! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!
!+
! Module: cheese_mod
!
! Authors:  Brendan Perry - NASA/GSFC, brendan.h.perry@nasa.gov
!            (Conversion to f95 from perl for SAS-19)
! K.D.Kuntz (NASA/GSFC) and The Johns Hopkins University, kuntz@pha.jhu.edu
!           Steven L. Snowden - NASA/GSFC (original perl algorithm)
! 
! Description: Extended Source Analysis Software
!
! Comments:
!
! $Id: cheese_mod.f90,v 1.3 2009/02/26 16:39:45 aibarra Exp $
!-
!  Version 0.11 - complete rewrite of sas task cheese.pl into f90 by BP
!
! This task creates "cheese" masks after running source detection on full-field
! images.  It uses the SAS tasks emask, region, and edetect_chain.
!
!  Usage:
!
module cheese_mod

  use dal
  use cal
  use param
  use dataFilter
  use errorHandling
  use sasdal
  use gofio_mod

  implicit none

  private

  character(len=*), parameter :: &
       mos1filespar = "mos1files", &   ! input EMOS event lists from espfilt
       mos2filespar = "mos2files", &   ! input EMOS event lists from espfilt
       pnfilespar   = "pnfiles", &     ! input EPN event lists from espfilt
       pnootpar     = "pnootfile", &   ! input EPN out of time event list
       scalepar     = "scale", &       ! E fraction (exclusion radius for pt srcs
       rateTpar     = "ratetotal", &   ! Flux threshold for exclusion of point sources
       rateSpar     = "ratesoft", &    ! Flux threshold for excl of pnt srcs (soft)
       rateHpar     = "ratehard", &    ! Flux threshold for excl of pnt srcs (hard)
       distpar      = "dist", &        ! Min sep (arcsecs) between masked sources
       mlminpar     = "mlmin", &       ! Min Max Likelihood detection value
       elowpar      = "elow", &        ! Low energy limit for the bands (eV)
       ehighpar     = "ehigh", &       ! High energy limit for the bands (eV)
       keepspar     = "keepinterfiles" ! keep (delete) intermediary files?
  
  public cheese

contains

  !---------------------------------------------------------------------
  subroutine cheese
    !---------------------------------------------------------------------

    type(DataSetT) :: inOBJIMG_S, inOBJSPEC_S, outSPdetmap_S, outSPskymap_S, outOBJIMG_S
    type(ArrayT)   :: inOBJIMG_A, outSPmap_A, out_A
    type(TableT)   :: inOBJSPEC_T, inOBJIMG_T
    character (len=2) :: inst, instO, instS, cc, bandd
    character (len=4) :: mos1ecf, mos2ecf, pnecf
    character (len=5) :: expid, any
    character (len=6) :: celow(2), cehigh(2)
    character (len=10) :: obsid, obsidO, obsidS, ctype(2), cunit(2), radecsys
    character (len=12) :: str_crval1, str_crval2, str_detx, str_dety
    character (len=18) :: prefix, type, suffix, logfile
    character (len=25) :: smode, smodeO, smodeS, submode
    character (len=256) :: inmos1_F(3), inmos2_F(3), inpn_F(3), inpnoot_F, &
         outM1fovT_F(3), outM1fovS_F(3), outM1fovH_F(3), outM2fovT_F(3), outM2fovS_F(3),&
         outM2fovH_F(3), outPNfovT_F(3), outPNfovS_F(3), outPNfovH_F(3)
    character (len=256) :: atthk_F, detmask_F, expimg_F, regiondet_F, regionsky_F, &
         cheese_F, piminlist, pimaxlist, ecflist
    character (len=1024) :: command, evtlist
    character (len=1024) :: imglist
    logical :: i_ok, warndie, mos1, mos2, pn, keep 
    integer, dimension(2) :: outIMGdim
    integer :: numMOS1ins, numMOS2ins, numPNins
    integer :: chansobs, instnum, typeF, nRowsO, nRowsS, revolut, revolutO, &
         revolutS, spectrol, inChans, spXsize, spYsize, ff
    integer :: elow(2), ehigh(2)
    integer :: i, j, k, n, nEVLIs
    integer :: status = 0, nBandsL, nBandsH
    integer(kind=int16), dimension(:,:), pointer :: aDI16
    integer(kind=int32), dimension(:,:), pointer :: outSPmap_IP, aDI32, &
         inOBJIMG_IP
    integer, parameter :: specchans = 2400
    integer, parameter :: xGood = 780, yGood = 780    
    real(single), dimension(:,:), pointer :: outSPmap_RP, flaremap_RP, aDR32
    real(double), dimension(:,:), pointer :: aDR64
    real(single), dimension(:,:), pointer :: temp_real32
    real :: tempp, dist, mlmin
    real :: eh, el, exposen, rn, scale, rateT, rateS, rateH, temp
    real(kind=double) :: crval(2), cdelt(2), crpix(2), crota(2), &
         dxcen, dycen, ra_pnt, dec_pnt, pa_pnt
    real :: backO, areaO, expoO, backS, areaS, expoS
    mos1ecf = '1.2' ! parameterize these???!!!
    mos2ecf = '1.2'
    pnecf   = '3.2'
    numMOS1ins = 0
    numMOS2ins = 0
    numPNins   = 0    
    numMOS1ins = parameterCount(mos1filespar)
    numMOS2ins = parameterCount(mos2filespar)
    numPNins   = parameterCount(pnfilespar)

    nEVLIs = numMOS1ins + numMOS2ins + numPNins
    if ((nEVLIs .gt. 3) .or. (nEVLIs .lt. 1)) then
       write(errstr,*) &
            "Total number of input event lists must be 1-3"
       call error("nEVLIlt1gt3",trim(errstr))
    endif
    !
    mos1 = .FALSE.
    mos2 = .FALSE.
    pn   = .FALSE.
    if (numMOS1ins .gt. 0) then
       mos1 = .TRUE.
    end if
    if (numMOS2ins .gt. 0) then
       mos2 = .TRUE.
    end if
    if (numPNins .gt. 0) then
       pn   = .TRUE.
    end if
    nBandsL = parameterCount(elowpar)
    nBandsH = parameterCount(ehighpar)
    celow = ''
    cehigh = ''
    if (nBandsL .gt. 2) then
       write(errstr,*) &
            "Number of elow values must be 1 or 2, exiting..."
       call error("eLowsGT2",trim(errstr))
    endif
    if (nBandsH .gt. 2) then
       write(errstr,*) &
            "Number of ehigh values must be 1 or 2, exiting..."
       call error("eHighsGT2",trim(errstr))
    endif
    if (nBandsL .ne. nBandsH) then
       write(errstr,*) &
            "Number of elow and ehigh values must be equal, exiting..."
       call error("eLowsNEeHighs",trim(errstr))
    else    
       do i=1,nBandsL
          elow(i)  = intParameter(elowpar,i-1)
          ehigh(i) = intParameter(ehighpar,i-1)
          write(celow(i),'(I5)') elow(i)
          write(cehigh(i),'(I5)') ehigh(i)
       end do
    end if
    if (nBandsL .eq. 1) then ! to make single "t" outputs
       celow(2) = celow(1)
       cehigh(2) = cehigh(1)
    endif
    !
    ! get the non-filename parameters
    !
    inpnoot_F = stringParameter(pnootpar)
    scale = realParameter(scalepar)
    rateT = realParameter(rateTpar)
    rateS = realParameter(rateSpar)
    rateH = realParameter(rateHpar)
    dist  = realParameter(distpar)
    mlmin = realParameter(mlminpar)
    keep  = booleanParameter(keepspar)
    if (verbosity(AppMsg) > Verbose) then
       call printParams(scale,rateT,rateS,rateH,dist,mlmin,keep)
    endif
    inmos1_F  = ""
    inmos2_F  = ""
    inpn_F    = ""
    imglist   = ""
    evtlist   = ""
    piminlist = ""
    pimaxlist = ""
    ecflist   = ""
    any       = ""
    !
    ! take infiles from params and make FOV images in one loop
    ! keep filenames in array for looping later
    ! build command line params for edetect_chain, emask, region, make_mask
    ! ADD CHECK_CLOBBER???!!!
    write(errstr,*) "Create filtered FOV image files, and edetect_chain params:"
    call message(AppMsg,Always)
    ! EMOS1
    if (mos1) then
       do i=1,numMOS1ins
          inmos1_F(i) = stringParameter(mos1filespar,i-1) ! mos1S001-allevc.fits OLD:fovevc
          ! total band first
          call makeOutName(inmos1_F(i),outM1fovT_F(i),"t",1,any) ! mos1S001-fovimt.fits
          call makeImFile(inmos1_F(i),outM1fovT_F(i),'M1',celow(1),cehigh(2))
          if (nBandsL .eq. 1) then
             imglist   = trim(adjustl(imglist)) // " " // trim(adjustl(outM1fovT_F(i)))
             evtlist   = trim(adjustl(evtlist)) // " " // trim(adjustl(inmos1_F(i)))
             piminlist = trim(adjustl(piminlist)) // " " // trim(celow(1))
             pimaxlist = trim(adjustl(pimaxlist)) // " " // trim(cehigh(1))
             ecflist   = trim(adjustl(ecflist)) // " " // mos1ecf
          elseif (nBandsL .eq. 2) then
             ! soft band
             call makeOutName(inmos1_F(i),outM1fovS_F(i),"s",1,any) ! mos1S001-fovims.fits
             call makeImFile(inmos1_F(i),outM1fovS_F(i),'M1',celow(1),cehigh(2))
             ! hard band
             call makeOutName(inmos1_F(i),outM1fovH_F(i),"h",1,any) ! mos1S001-fovimh.fits
             call makeImFile(inmos1_F(i),outM1fovH_F(i),'M1',celow(2),cehigh(2))
             imglist   = trim(adjustl(imglist)) // " " // &
                  trim(adjustl(outM1fovS_F(i))) // " " // &
                  trim(adjustl(outM1fovH_F(i))) // " " 
             evtlist   = trim(adjustl(evtlist)) // " " // trim(adjustl(inmos1_F(i)))
             piminlist = trim(adjustl(piminlist)) // " " // &
                  trim(adjustl(celow(1))) // " " // trim(adjustl(celow(2)))
             pimaxlist = trim(adjustl(pimaxlist)) // " " // &
                  trim(adjustl(cehigh(1))) // " " // trim(adjustl(cehigh(2)))
             ecflist = trim(adjustl(ecflist)) // " " // mos1ecf // mos1ecf
          end if
       end do
    end if
    ! EMOS2
    if (mos2) then
       do i=1,numMOS2ins
          inmos2_F(i) = stringParameter(mos2filespar,i-1) ! mos2S002-fovevc.fits
          ! total band first
          call makeOutName(inmos2_F(i),outM2fovT_F(i),"t",1,any) ! mos2S001-fovimt.fits
          call makeImFile(inmos2_F(i),outM2fovT_F(i),'M2',celow(1),cehigh(2))
          if (nBandsL .eq. 1) then
             imglist   = trim(adjustl(imglist)) // " " // trim(adjustl(outM2fovT_F(i)))
             evtlist   = trim(adjustl(evtlist)) // " " // trim(adjustl(inmos2_F(i)))
             piminlist = trim(adjustl(piminlist)) // " " // trim(celow(1))
             pimaxlist = trim(adjustl(pimaxlist)) // " " // trim(cehigh(1))
             ecflist   = trim(adjustl(ecflist)) // " " // mos2ecf
          elseif (nBandsL .eq. 2) then
             ! soft band
             call makeOutName(inmos2_F(i),outM2fovS_F(i),"s",1,any) ! mos2S002-fovims.fits
             call makeImFile(inmos2_F(i),outM2fovS_F(i),'M2',celow(1),cehigh(2))
             ! hard band
             call makeOutName(inmos2_F(i),outM2fovH_F(i),"h",1,any) ! mos2S002-fovimh.fits
             call makeImFile(inmos2_F(i),outM2fovH_F(i),'M2',celow(2),cehigh(2))
             imglist   = trim(adjustl(imglist)) // " " // &
                  trim(adjustl(outM2fovS_F(i))) // " " // &
                  trim(adjustl(outM2fovH_F(i))) // " " 
             evtlist   = trim(adjustl(evtlist)) // " " // trim(adjustl(inmos2_F(i)))
             piminlist = trim(adjustl(piminlist)) // " " // &
                  trim(adjustl(celow(1))) // " " // trim(adjustl(celow(2)))
             pimaxlist = trim(adjustl(pimaxlist)) // " " // &
                  trim(adjustl(cehigh(1))) // " " // trim(adjustl(cehigh(2)))
             ecflist   = trim(adjustl(ecflist)) // " " // mos2ecf // mos2ecf
          end if
       end do
    end if
    ! EPN
    if (pn) then
       do i=1,numPNins
          inpn_F(i) = stringParameter(pnfilespar,i-1)   ! pnS001-fovevc.fits
          ! total band first
          call makeOutName(inpn_F(i),outPNfovT_F(i),"t",1,any) ! pnS001-fovimt.fits
          call makeImFile(inpn_F(i),outPNfovT_F(i),'PN',celow(1),cehigh(2))
          if (nBandsL .eq. 1) then
             imglist   = trim(adjustl(imglist)) // " " // trim(adjustl(outPNfovT_F(i)))
             evtlist   = trim(adjustl(evtlist)) // " " // trim(adjustl(inpn_F(i)))
             piminlist = trim(adjustl(piminlist)) // " " // trim(celow(1))
             pimaxlist = trim(adjustl(pimaxlist)) // " " // trim(cehigh(1))
             ecflist   = trim(adjustl(ecflist)) // " " // pnecf
          elseif (nBandsL .eq. 2) then
             ! soft band
             call makeOutName(inpn_F(i),outPNfovS_F(i),"s",1,any) ! pnS001-fovims.fits
             call makeImFile(inpn_F(i),outPNfovS_F(i),'PN',celow(1),cehigh(2))
             ! hard band
             call makeOutName(inpn_F(i),outPNfovH_F(i),"h",1,any) ! pnS001-fovimh.fits
             call makeImFile(inpn_F(i),outPNfovH_F(i),'PN',celow(2),cehigh(2))
             imglist   = trim(adjustl(imglist)) // " " // &
                  trim(adjustl(outPNfovS_F(i))) // " " // &
                  trim(adjustl(outPNfovH_F(i))) // " " 
             evtlist   = trim(adjustl(evtlist)) // " " // trim(adjustl(inpn_F(i)))
             piminlist = trim(adjustl(piminlist)) // " " // &
                  trim(adjustl(celow(1))) // " " // trim(adjustl(celow(2)))
             pimaxlist = trim(adjustl(pimaxlist)) // " " // &
                  trim(adjustl(cehigh(1))) // " " // trim(adjustl(cehigh(2)))
             ecflist   = trim(adjustl(ecflist)) // " " // pnecf // pnecf
          end if
       end do
    end if
    ! clear out superfluous white space (though edetect_chain works anyway)???!!!
    write(errstr,*) "evtlist   = ",trim(adjustl(evtlist))
    call message(AppMsg,Always)
    write(errstr,*) "imglist   = ",trim(adjustl(imglist))
    call message(AppMsg,Always)
    write(errstr,*) "piminlist = ",trim(adjustl(piminlist))
    call message(AppMsg,Always)
    write(errstr,*) "pimaxlist = ",trim(adjustl(pimaxlist))
    call message(AppMsg,Always)
    write(errstr,*) "ecflist   = ",trim(adjustl(ecflist))
    call message(AppMsg,Always)
    !
    ! check for atthk.fits, else run atthkgen to make attitude file
    !
    atthk_F = 'atthk.fits'
    inquire(FILE=trim(atthk_F), EXIST=i_ok)
    if (i_ok) then
       if ( clobber() ) then
          write(errstr,*) "Attitude file: ",trim(atthk_F),&
               " exists and CLOBBER is set, will overwrite existing file."
          call message(AppMsg,Always)          
          call unlink (trim(atthk_F))
          command = "atthkgen atthkset="// adjustl(trim(atthk_F)) //" timestep=1 "
          write(errstr,*) "*atthkgen command = "
          call message(AppMsg,Always)
          write(errstr,*) trim(adjustl(command))
          call message(AppMsg,Always)
          call system(trim(command))
       else
          write(errstr,*) "Attitude file ",trim(adjustl(atthk_F)),&
               " exists and CLOBBER is NOT set, using existing..."
          call message(AppMsg,Always)
       endif
    else
       write(errstr,*) "Attitude file: ",trim(atthk_F),&
            " does not exist, running atthkgen"
       call message(AppMsg,Always)
       command = "atthkgen atthkset="// adjustl(trim(atthk_F)) //" timestep=1 "
       write(errstr,*) "*atthkgen command = "
       call message(AppMsg,Always)
       write(errstr,*) trim(adjustl(command))
       call message(AppMsg,Always)
       call system(trim(command))
    end if
    !
    !
    call run_edetectchain(evtlist,imglist,piminlist,pimaxlist,ecflist,inpnoot_F)
    !
    ! Now make region files, masks, and cheese masks
    ! EMOS1
    if (mos1) then
       do i=1,numMOS1ins
          ! total band
          call makeOutName(inmos1_F(i),detmask_F,"t",1,"mask") ! mos1S001-fovimtmask
          call makeOutName(inmos1_F(i),expimg_F,"t",1,"exp")   ! mos1S001-fovimtexp
          call run_emask(detmask_F,expimg_F)
          call makeOutName(inmos1_F(i),regiondet_F,"t",2,any) ! mos1S001-bkgregdett.fits
          call makeOutName(inmos1_F(i),regionsky_F,"t",3,any) ! mos1S001-bkgregskyt.fits
          call run_region(inmos1_F(i),regiondet_F,regionsky_F,ratet,mlmin,2,&
               dist,scale,1)
          call makeOutName(inmos1_F(i),cheese_F,"t",5,any) ! mos1S001-cheeset.fits
          call run_makemask(outM1fovT_F(i),detmask_F,regionsky_F,cheese_F)
          if (nBandsL .gt. 1) then
             ! soft band
             call makeOutName(inmos1_F(i),detmask_F,"s",1,"mask") ! mos1S001-fovimsmask
             call makeOutName(inmos1_F(i),expimg_F,"s",1,"exp")   ! mos1S001-fovimsexp
             call run_emask(detmask_F,expimg_F)
             call makeOutName(inmos1_F(i),regiondet_F,"s",2,any) ! mos1S001-bkgregdets.fits
             call makeOutName(inmos1_F(i),regionsky_F,"s",3,any) ! mos1S001-bkgregskys.fits
             call run_region(inmos1_F(i),regiondet_F,regionsky_F,rates,mlmin,2,&
                  dist,scale,1)
             call makeOutName(inmos1_F(i),cheese_F,"s",5,any) ! mos1S001-cheeses.fits
             call run_makemask(outM1fovS_F(i),detmask_F,regionsky_F,cheese_F)
             ! hard band
             call makeOutName(inmos1_F(i),detmask_F,"h",1,"mask") ! mos1S001-fovimsmask
             call makeOutName(inmos1_F(i),expimg_F,"h",1,"exp")   ! mos1S001-fovimsexp
             call run_emask(detmask_F,expimg_F)
             call makeOutName(inmos1_F(i),regiondet_F,"h",2,any) ! mos1S001-bkgregdeth.fits
             call makeOutName(inmos1_F(i),regionsky_F,"h",3,any) ! mos1S001-bkgregskyh.fits
             call run_region(inmos1_F(i),regiondet_F,regionsky_F,rateh,mlmin,2,&
                  dist,scale,2)
             call makeOutName(inmos1_F(i),cheese_F,"h",5,any) ! mos1S001-cheeseh.fits
             call run_makemask(outM1fovH_F(i),detmask_F,regionsky_F,cheese_F)
          end if
       end do
    end if
    ! EMOS2
    if (mos2) then
       do i=1,numMOS2ins
          ! total band 
          call makeOutName(inmos2_F(i),detmask_F,"t",1,"mask") ! mos2S002-fovimtmask
          call makeOutName(inmos2_F(i),expimg_F,"t",1,"exp")   ! mos2S002-fovimtexp
          call run_emask(detmask_F,expimg_F)
          call makeOutName(inmos2_F(i),regiondet_F,"t",2,any) ! mos2S002-bkgregdett.fits
          call makeOutName(inmos2_F(i),regionsky_F,"t",3,any) ! mos2S001-bkgregskyt.fits
          call run_region(inmos2_F(i),regiondet_F,regionsky_F,ratet,mlmin,3,&
               dist,scale,1)
          call makeOutName(inmos2_F(i),cheese_F,"t",5,any) ! mos2S002-tcheese.fits
          call run_makemask(outM2fovT_F(i),detmask_F,regionsky_F,cheese_F)
          if (nBandsL .gt. 1) then
             ! soft band
             call makeOutName(inmos2_F(i),detmask_F,"s",1,"mask") ! mos2S002-fovimsmask
             call makeOutName(inmos2_F(i),expimg_F,"s",1,"exp")   ! mos2S002-fovimsexp
             call run_emask(detmask_F,expimg_F)
             call makeOutName(inmos2_F(i),regiondet_F,"s",2,any) ! mos1S001-bkgregdets.fits
             call makeOutName(inmos2_F(i),regionsky_F,"s",3,any) ! mos1S001-bkregdets.fits
             call run_region(inmos2_F(i),regiondet_F,regionsky_F,rates,mlmin,3,&
                  dist,scale,1)
             call makeOutName(inmos2_F(i),cheese_F,"s",5,any) ! mos2S002-scheese.fits
             call run_makemask(outM2fovS_F(i),detmask_F,regionsky_F,cheese_F)
             ! hard band
             call makeOutName(inmos2_F(i),detmask_F,"h",1,"mask") ! mos2S002-fovimhmask
             call makeOutName(inmos2_F(i),expimg_F,"h",1,"exp")   ! mos2S002-fovimhexp
             call run_emask(detmask_F,expimg_F)
             call makeOutName(inmos2_F(i),regiondet_F,"h",2,any) ! mos1S001-bkregdeth.fits
             call makeOutName(inmos2_F(i),regionsky_F,"h",3,any) ! mos1S001-bkregdeth.fits
             call run_region(inmos2_F(i),regiondet_F,regionsky_F,rateh,mlmin,3,&
                  dist,scale,2)
             call makeOutName(inmos2_F(i),cheese_F,"h",5,any) ! mos2S002-hcheese.fits
             call run_makemask(outM2fovH_F(i),detmask_F,regionsky_F,cheese_F)
          end if
       end do
    end if
    ! PN
    if (pn) then
       do i=1,numPNins
          ! total band
          call makeOutName(inpn_F(i),detmask_F,"t",1,"mask") ! pnS002-fovimtmask
          call makeOutName(inpn_F(i),expimg_F,"t",1,"exp")   ! pnS002-fovimtexp
          call run_emask(detmask_F,expimg_F)
          call makeOutName(inpn_F(i),regiondet_F,"t",2,any) ! mos1S001-bkregdett.fits
          call makeOutName(inpn_F(i),regionsky_F,"t",3,any) ! mos1S001-bkregdett.fits
          call run_region(inpn_F(i),regiondet_F,regionsky_F,ratet,mlmin,3,&
               dist,scale,1)
          call makeOutName(inpn_F(i),cheese_F,"t",5,any) ! pnS003-tcheese.fits
          call run_makemask(outPNfovT_F(i),detmask_F,regionsky_F,cheese_F)
          if (nBandsL .gt. 1) then
             ! soft band
             call makeOutName(inpn_F(i),detmask_F,"s",1,"mask") ! pnS002-fovimsmask
             call makeOutName(inpn_F(i),expimg_F,"s",1,"exp")   ! pnS002-fovimsexp
             call run_emask(detmask_F,expimg_F)
             call makeOutName(inpn_F(i),regiondet_F,"s",2,any) ! mos1S001-bkregdets.fits
             call makeOutName(inpn_F(i),regionsky_F,"s",3,any) ! mos1S001-bkregdets.fits
             call run_region(inpn_F(i),regiondet_F,regionsky_F,rates,mlmin,3,&
                  dist,scale,1)
             call makeOutName(inpn_F(i),cheese_F,"s",5,any) ! pnS003-scheese.fits
             call run_makemask(outPNfovS_F(i),detmask_F,regionsky_F,cheese_F)
             ! hard band
             call makeOutName(inpn_F(i),detmask_F,"h",1,"mask") ! pnS002-fovimhmask
             call makeOutName(inpn_F(i),expimg_F,"h",1,"exp")   ! pnS002-fovimhexp
             call run_emask(detmask_F,expimg_F)
             call makeOutName(inpn_F(i),regiondet_F,"h",2,any) ! mos1S001-bkregdeth.fits
             call makeOutName(inpn_F(i),regionsky_F,"h",3,any) ! mos1S001-bkregdeth.fits
             call run_region(inpn_F(i),regiondet_F,regionsky_F,rateh,mlmin,3,&
                  dist,scale,2)
             call makeOutName(inpn_F(i),cheese_F,"h",5,any) ! pnS003-hcheese.fits
             call run_makemask(outPNfovH_F(i),detmask_F,regionsky_F,cheese_F)
          end if
       end do
    end if
    ! which of these can be deleted???!!!
    ! cheese makes these files (plus s and h versions if bands selected):
    ! pnS003-bkregtsky.fits, pnS003-cheeset.fits, mos2S002-cheeset.fits,
    ! pnS003-bkregtdet.fits,pnS003-fovimtmask.fits,mos1S001-bkregtsky.fits,
    ! mos1S001-cheeset.fits,mos2S002-bkregtdet.fits,mos2S002-bkregtsky.fits,
    ! mos2S002-fovimtmask.fits,mos1S001-bkregtdet.fits,mos1S001-fovimtmask.fits,
    ! mos2S002-fovimtsen.fits,mos1S001-fovimtsen.fits,pnS003-fovimtsen.fits,
    ! emllist.fits,mos1S001-fovimtsmap.fits,mos2S002-fovimtsmap.fits,
    ! pnS003-fovimtsmap.fits,eboxlist_m.fits,mos2S002-fovimtbkg.fits,
    ! mos1S001-fovimtbkg.fits,pnS003-fovimtbkg.fits,eboxlist_l.fits,
    ! mos2S002-fovimtexp.fits,mos1S001-fovimtexp.fits,pnS003-fovimtexp.fits,
    ! atthk.fits,filtered.fits,pnS003-fovimt.fits,mos1S001-fovimt.fits,
    ! mos2S002-fovimt.fits
    !
    if (.not.keep) then
       ! add actual files to kill here ???!!!
       write(errstr,*) "CHEESE: keepinterfiles NOT set"
       call message(AppMsg,Always)
       write(errstr,*) "CHEESE: Deleting intermediary files..."
       call message(AppMsg,Always)
       !write(errstr,*) "Object image: ",trim(adjustl(objimg_F))
       !call message(AppMsg,Always)
       !call unlink(trim(adjustl(objimg_F)))
    end if

    return

  end subroutine cheese

  subroutine makeOutName(infile,outfile,band,type,any)
    !
    ! create output file name based on input eventlist
    ! 
    type(DataSetT) :: in_S
    type(TableT)   :: in_T
    type(ArrayT)   :: in_A
    character(len=*), intent(in) :: any
    character(len=1) :: band
    character(len=2) :: inst
    character(len=6) :: instr, filter, coord, suffix
    character(len=10) :: obsid, expid, brand
    character(len=25) :: smode
    character(len=256), intent(in) :: infile
    character(len=256), intent(out) :: outfile
    integer :: typeF, status, revolut, nRows, type, frmtime, &
         elow, ehigh
    logical :: warndie
    real :: back, area, expo
    suffix = '.fits' ! default case
    coord = ''
    brand = ''
    select case (type)
    case(1) ! fov image in sky coords
       brand   = "-fovim"
    case(2) ! region file in det coords
       brand   = "-bkgreg" ! bkregtreg
       coord  = "det" 
    case(3) ! region file in sky coords
       brand   = "-bkgreg" ! bkregtreg
       coord  = "sky" 
    case(4) ! emask file
       brand = "-fovexp"
    case(5) ! final cheese mask
       brand = "-cheese"
    case default
       brand = ''
    end select
    write(errstr,*) &
         "CHEESE: Reading info from this file to make new name: ",trim(infile)
    call message(AppMsg,Always)
    status = 0
    warndie = .TRUE.
    call check_open_set(trim(infile),in_S,1,status,warndie)
    !
    ! Get critical header keywords 
    !
    status = 0
    warndie = .TRUE.
    typeF = 1 ! 1 = EVLI, 2 = SPEC, 3 = IMG, most of these are dummies
    call read_header_info(typeF,in_S,in_T,in_A,inst,obsid,expid,&
         smode,frmtime,filter,revolut,back,area,expo,elow,ehigh,&
         nRows,warndie)
    select case (inst)
    case('M1') 
       instr = "mos1"
    case('M2')
       instr = "mos2"
    case('PN')
       instr = "pn"
    case default
       instr = ""
    end select
    write(errstr,*) "CHEESE: Old (input) file     : ",trim(infile)
    call message(AppMsg,Always)
    outfile = trim(adjustl(instr)) // trim(adjustl(expid)) // &
         trim(adjustl(brand)) // trim(adjustl(band)) // &
         trim(adjustl(coord)) // trim(adjustl(any)) // &
         trim(adjustl(suffix))
    write(errstr,*) "CHEESE: New (output) filename: ",trim(outfile)
    call message(AppMsg,Always)
    return
    
  end subroutine makeOutName
  !
  subroutine makeImFile(infile,outfile,inst,celow,cehigh)
    !
    ! Make the clean FOV image files in sky coords
    !
    character(len=*), intent(inout) :: infile, outfile
    character(len=*), intent(in) :: celow, cehigh
    character(len=2) :: inst
    character(len=1024) :: command
    integer :: nMOS1, nMOS2, nPN, elow, ehigh
    logical :: i_ok
    
    if ((inst .eq. 'M1') .or. (inst .eq. 'M2')) then 
       write(errstr,*) "Making FOV sky coord image files for " // inst
       call message(AppMsg,Always)
       inquire(FILE=trim(outfile), EXIST=i_ok)
       if (i_ok) then
          if ( clobber() ) then
             write(errstr,*) inst // " FOV image file: ",trim(outfile),&
                  " exists and CLOBBER is set, will overwrite existing file."
             call message(AppMsg,Always)          
             call unlink (trim(outfile))
          else
             write(errstr,*) inst // " FOV image file: ",trim(outfile),&
                  " exists and CLOBBER is NOT set, using existing..."
             call message(AppMsg,Always)
          endif
       end if
       command =  "evselect table="// trim(adjustl(infile)) //":EVENTS " // &
            "withfilteredset=yes expression='(PATTERN<=12)&&((FLAG & 0x766aa000) == 0)&&" // &
            "(PI in [" // trim(adjustl(celow)) // ":" // trim(adjustl(cehigh)) // &
            "])' " // "filtertype=expression imagebinning='imageSize' " // &
            "imagedatatype='Int32' imageset=" // trim(adjustl(outfile)) // &
            " squarepixels=yes ignorelegallimits=yes "// &
            "withxranges=yes withyranges=yes xcolumn='X' ximagesize=900 " // &
            "ximagemax=48400 ximagemin=3401 ycolumn='Y' yimagesize=900 " // &
            "yimagemax=48400 yimagemin=3401 updateexposure=yes filterexposure=yes " // &
            "verbosity=4"
       write(errstr,*) inst // " FOV image evselect command = "
       call message(AppMsg,Always)
       write(errstr,*) trim(adjustl(command))
       call message(AppMsg,Always)
       call system(trim(command))
    !
    elseif (inst .eq. 'PN') then 
       write(errstr,*) "Making FOV sky coord image files for PN"
       call message(AppMsg,Always)
       inquire(FILE=trim(outfile), EXIST=i_ok)
       if (i_ok) then
          if ( clobber() ) then
             write(errstr,*) inst // " FOV image file: ",trim(outfile),&
                  " exists and CLOBBER is set, will overwrite existing file."
             call message(AppMsg,Always)          
             call unlink (trim(outfile))
          else
             write(errstr,*) inst // " FOV image file: ",trim(outfile),&
                  " exists and CLOBBER is NOT set, using existing..."
             call message(AppMsg,Always)
          endif
       end if
       command = "evselect table="// trim(adjustl(infile)) // ":EVENTS " //&
            "withfilteredset=yes expression='(PATTERN<=4)&&((FLAG & 0x762a097c) == 0)&&" //&
            "(PI in [" // trim(adjustl(celow)) // ":" // trim(adjustl(cehigh)) // &
            "])&&" // "(DETY in [-16510:14345])' filtertype=expression " //& 
            "imagebinning='imageSize' imagedatatype='Int32' " //&
            "imageset=" // trim(adjustl(outfile)) //&
            " ignorelegallimits=yes withxranges=yes withyranges=yes " //&
            "xcolumn='X' ximagesize=900 ximagemax=48400 ximagemin=3401 " //&
            "ycolumn='Y' yimagesize=900 yimagemax=48400 yimagemin=3401 " //& 
            "updateexposure=yes filterexposure=yes verbosity=4"
       write(errstr,*) "*PN Flare free FOV image evselect command = "
       write(errstr,*) inst // " Flare free FOV image evselect command = "
       call message(AppMsg,Always)
       write(errstr,*) trim(adjustl(command))
       call message(AppMsg,Always)
       call system(trim(command))
    else
       call error("notM1M2PN","Instrument is not M1, M2, or PN, weirdly")
    end if
    
    return
    
  end subroutine makeImFile
  !
  !
  !
  subroutine run_edetectchain(evtlist,imglist,piminlist,pimaxlist,ecflist,pnoot_F)
    !
    ! routine to run SAS task edetect_chain
    !
    character(len=1024), intent(in) :: imglist, evtlist
    character(len=1024) :: command
    character(len=256), intent(in) :: piminlist, pimaxlist, ecflist, pnoot_F
    character(len=12) :: emllist_F = "emllist.fits"
    logical :: i_ok
    
    inquire(FILE=trim(emllist_F), EXIST=i_ok)
    if (i_ok) then
       if ( clobber() ) then
          write(errstr,*) " emllist file: ",trim(emllist_F),&
               " exists and CLOBBER is set, will overwrite existing file."
          call message(AppMsg,Always)          
          call unlink (trim(emllist_F))
       else
          write(errstr,*) " emllist file: ",trim(emllist_F),&
               " exists and CLOBBER is NOT set, using existing..."
          call message(AppMsg,Always)
          return
       endif
    end if
    !
    if (pnoot_F .ne. "") then
       command = "edetect_chain imagesets='" // trim(adjustl(imglist)) // &
            " ' eventsets='" // trim(adjustl(evtlist)) // &
            " ' attitudeset=atthk.fits pimin='" // trim(adjustl(piminlist)) // &
            " ' pimax='" // trim(adjustl(pimaxlist)) // &
            " ' witheexpmap=yes likemin=10 ecf='" // trim(adjustl(ecflist)) // &
            " ' eboxl_likemin=15 eboxm_likemin=15 ebox_withdetmask=yes "// &
            " ebox_withexpimage=yes esp_scut=0.01 esp_withootset=yes " // &
            " esp_ooteventset='" // trim(adjustl(pnoot_F)) // "' esp_nsplinenodes=20" // &
            " eml_ecut=15 eml_scut=15 eml_withtwostage=yes eml_withthreshold=yes " // &
            " eml_threshold=15 esen_mlmin=10 eml_list='emllist.fits'"
    else
       command = "edetect_chain imagesets='" // trim(adjustl(imglist)) // &
            " ' eventsets='" // trim(adjustl(evtlist)) // &
            " ' attitudeset=atthk.fits pimin='" // trim(adjustl(piminlist)) // &
            " ' pimax='" // trim(adjustl(pimaxlist)) // &
            " ' witheexpmap=yes likemin=10 ecf='" // trim(adjustl(ecflist)) // &
            " ' eboxl_likemin=15 eboxm_likemin=15 ebox_withdetmask=yes "// &
            " ebox_withexpimage=yes esp_scut=0.01 esp_withootset=no " // &
            " esp_nsplinenodes=20" // &
            " eml_ecut=15 eml_scut=15 eml_withtwostage=yes eml_withthreshold=yes " // &
            " eml_threshold=15 esen_mlmin=10 eml_list='emllist.fits'"
    end if

    write(errstr,*) "*edetect_chain command = "
    call message(AppMsg,Always)
    write(errstr,*) trim(adjustl(command))
    call message(AppMsg,Always)
    call system(trim(command))
    return
    
  end subroutine run_edetectchain
  !
  !
  !
  subroutine run_region(infile,detfile,skyfile,rate,mlmin,det,dist,scale,idband)
    !
    ! Make background region files
    !
    character(len=*), intent(in) :: infile, detfile, skyfile
    character(len=10) :: cdist, crate, cdet, cmlmin, cscale, cidband
    character(len=1024) :: command
    integer :: det, idband
    real :: rate, mlmin, dist, scale

    ! formats taken from parameter constraints. 
    write(cdet,'(I1)') det
    write(cidband,'(I1)') idband
    write(cdist,'(F7.4)') dist
    write(crate,'(F7.3)') rate 
    write(cmlmin,'(F7.4)') mlmin
    write(cscale,'(F7.4)') scale
    crate = trim(adjustl(crate)) // "e-14"
    !           
    command = "region eventset=" // trim(adjustl(infile)) // &
         " operationstyle=global srclisttab=emllist.fits:SRCLIST " // & 
         "expression='(DIST_NN >= " // trim(adjustl(cdist)) // &
         ")&&(FLUX >= " //  trim(adjustl(crate)) // ")" // &
         "&&(ID_INST == " // trim(adjustl(cdet)) // ")&&(DET_ML >= " // &
         trim(adjustl(cmlmin)) // ")&&(ID_BAND == " // trim(adjustl(cidband)) // &
         ")' bkgregionset=" // trim(adjustl(detfile)) // &
         " bkgfraction=" // trim(adjustl(cscale)) // &
         " radiusstyle=contour nosrcellipse=no " // & 
         "outunit=detxy verbosity=4"
    write(errstr,*) " Running region in DET coords command = "
    call message(AppMsg,Always)
    write(errstr,*) trim(adjustl(command))
    call message(AppMsg,Always)
    call system(trim(command))
    command = "region eventset=" // trim(adjustl(infile)) // &
         " operationstyle=global srclisttab=emllist.fits:SRCLIST " // & 
         "expression='(DIST_NN >= " // trim(adjustl(cdist)) // &
         ")&&(FLUX >= " //  trim(adjustl(crate)) // ")" // &
         "&&(ID_INST == " // trim(adjustl(cdet)) // ")&&(DET_ML >= " // &
         trim(adjustl(cmlmin)) // ")&&(ID_BAND == " // trim(adjustl(cidband)) // &
         ")' bkgregionset=" // trim(adjustl(skyfile)) // &
         " bkgfraction=" // trim(adjustl(cscale)) // &
         " radiusstyle=contour nosrcellipse=no " // & 
         "outunit=xy verbosity=4"
    write(errstr,*) " Running region in SKY coords command = "
    call message(AppMsg,Always)
    write(errstr,*) trim(adjustl(command))
    call message(AppMsg,Always)
    call system(trim(command))
    !    
    return

  end subroutine run_region

  subroutine run_emask(infile,maskfile)
    !
    ! Make masks files
    !
    character(len=*), intent(in) :: infile, maskfile
    character(len=1024) :: command
    command = "emask detmaskset=" // trim(adjustl(infile)) // & 
	  " expimageset=" // trim(adjustl(maskfile)) // & 
	  " threshold1=0.1 threshold2=0.5 verbosity=4"
    write(errstr,*) " Running emask command = "
    call message(AppMsg,Always)
    write(errstr,*) trim(adjustl(command))
    call message(AppMsg,Always)
    call system(trim(command))
    
    return

  end subroutine run_emask
  !
  !
  !
  subroutine run_makemask(imagefile,maskfile,regionfile,cheesefile)
    !
    ! Make cheese masks files
    ! imagefile : old=mos1S001-obj-im.fits  new=mos1S001-fovimt.fits
    ! maskfile  : old=mos1S001-mask-im.fits new=mos1S001-fovimtmask.fits
    ! cheesefile: old=mos1S001-cheese.fits  new=mos1S001-fovimtcheese.fits
    ! regionfile: old=mos1S001-bkg_region-sky.fits new=mos1S001-bkgregsky.fits
    !
    character(len=*), intent(in) :: imagefile, maskfile, cheesefile, regionfile
    character(len=1024) :: command
    ! in
    command = "makemask imagefile=" // trim(adjustl(imagefile)) // &
         " maskfile=" // trim(adjustl(maskfile)) // &
         " cheesefile=" // trim(adjustl(cheesefile)) // &
         " regionfile=" // trim(adjustl(regionfile))
    
    write(errstr,*) " Running makemask command = "
    call message(AppMsg,Always)
    write(errstr,*) trim(adjustl(command))
    call message(AppMsg,Always)
    call system(trim(command))
    
    return
    
  end subroutine run_makemask
  !
  !
  !
  subroutine printparams (scale,rateT,rateS,rateH,dist,mlmin,keep)

    real :: scale, rateT, rateS, rateH, dist, mlmin
    logical :: keep

    write(errstr,*) "Energy Fraction (exclusion radius for point sources):     ",scale
    call message(AppMsg,Always)
    write(errstr,*) "Flux threshold for exclusion of point sources (fullband): ",rateT
    call message(AppMsg,Always)
    write(errstr,*) "Flux threshold for exclusion of point sources (softband): ",rateS
    call message(AppMsg,Always)
    write(errstr,*) "Flux threshold for exclusion of point sources (hardband): ",rateH
    call message(AppMsg,Always)
    write(errstr,*) " Minimum separation (arcsec) between masked sources:      ",dist
    call message(AppMsg,Always)
    write(errstr,*) "Minimum Maximum Likelihood detection value:               ",mlmin
    call message(AppMsg,Always)
    write(errstr,*) "Keep intermediary files?: ",keep
    call message(AppMsg,Always)
    return

  end subroutine printparams

  subroutine catch_nfiles(mos1,mos2,pn,nM1,nM2,nPN)
    !
    ! number of other input files must be equal unless other is zero
    !

    logical :: mos1, mos2, pn
    integer :: nM1, nM2, nPN
    ! this catches all combos, right???!!!
    if (mos1) then
       if (mos2) then ! if both, they have to be equal
          if ((nM1 .ne. nM2) .and. (nM2 .ne. 0)) then
             write(errstr,*) &
                  "Number of MOS1 and MOS2 input files must be equal, exiting..."
             call error("MOS1neMOS2",trim(errstr))
          endif
       endif
       if (pn) then ! if only M1 and PN, they have to be equal
          if (nPN .ne. nM1) then
             write(errstr,*) &
                  "Number of MOS1 and PN input files must be equal, exiting..."
             call error("MOS1nePN",trim(errstr))
          endif
       end if
    end if
    if (mos2) then
       if (pn) then ! if both, they have to be equal
          if ((nM2 .ne. nPN) .and. (nPN .ne. 0)) then
             write(errstr,*) &
                  "Number of MOS2 and PN input files must be equal, exiting..."
             call error("MOS2nePN",trim(errstr))
          endif
       endif
    end if
    if ((.not. mos1) .and. (.not. mos2) .and. (.not. pn)) then
       ! catch non-mandatoryness of params
       write(errstr,*) &
            "mos1files/mos2files/pnfiles param not mandatory, but need at least one"
       call error("noMOS1noMOS2noPN",trim(errstr))
    end if
    return
  end subroutine catch_nfiles

  
end module cheese_mod
